import tensorflow as tf
import numpy as np
from .misc import periodic_distance, variable_summaries, add_layer_summaries, distance_cost
import os
from .parameters import Parameters
from tqdm import tqdm


class Autoencoder:

    def __init__(self, parameters, train_data=None, validation_data=None, checkpoint_path=None, n_inputs=None):
        """
        :param parameters: Parameters object as defined in :py:class:`encodermap.parameters.Parameters`

        :param train_data: 2d numpy array where each row is treated as a training point

        :param validation_data: A 2d numpy array. This data will only be used to calculate a validation error during
                                training. It will not be used for training.

        :param checkpoint_path: If a checkpoint path is given values like neural network weights stored in this
                                checkpoint will be restored.

        :param n_inputs: If no train_data is given, for example when an already trained network is restored from a
                         checkpoint, the number of of inputs needs to be given. This should be equal to the number of
                         columns of the train_data the network was originally trained with.
        """
        # Parameters:
        self.p = parameters  # type: Parameters
        self.p.save()
        print("Output files are saved to {}".format(self.p.main_path),
              "as defined in 'main_path' in the parameters.")

        self.graph = tf.Graph()
        with self.graph.as_default():

            # Load Data:
            if train_data is None:
                assert n_inputs is not None, "If no train_data is given, n_inputs needs to be given"
                self.train_data = np.zeros((3, n_inputs), dtype=np.float32)
            else:
                self.train_data = train_data.astype(np.float32)

            if validation_data is None:
                self.validation_data = None
            else:
                self.validation_data = validation_data.astype(np.float32)

            self.data_placeholder = tf.placeholder(self.train_data.dtype, self.train_data.shape)
            self.data_set = tf.data.Dataset.from_tensor_slices(self.data_placeholder)
            self.data_set = self.data_set.shuffle(buffer_size=len(self.train_data))
            self.data_set = self.data_set.repeat()
            self.data_set = self.data_set.batch(self.p.batch_size)
            self.data_iterator = self.data_set.make_initializable_iterator()

            # Setup Network:
            self.inputs = self.data_iterator.get_next()
            self.inputs = tf.placeholder_with_default(self.inputs, self.inputs.shape)
            self.regularizer = tf.contrib.layers.l2_regularizer(scale=self.p.l2_reg_constant)
            encoded = self._encode(self.inputs)
            self.latent = tf.placeholder_with_default(encoded, encoded.shape)
            variable_summaries("latent", self.latent)
            self.generated = self._generate(self.latent)

            # Define Cost function:
            self.cost = self._cost()

            # Setup Optimizer:
            self.optimizer = tf.train.AdamOptimizer(self.p.learning_rate)
            gradients = self.optimizer.compute_gradients(self.cost)
            self.optimize = self.optimizer.apply_gradients(gradients)

            self.merged_summaries = tf.summary.merge_all()

            # Setup Session
            gpu_options = tf.GPUOptions(per_process_gpu_memory_fraction=self.p.gpu_memory_fraction)
            self.sess = tf.Session(config=tf.ConfigProto(gpu_options=gpu_options), graph=self.graph)
            self.sess.run(tf.global_variables_initializer())
            self.sess.run(self.data_iterator.initializer, feed_dict={self.data_placeholder: self.train_data})
            self.train_writer = tf.summary.FileWriter(os.path.join(self.p.main_path, "train"), self.sess.graph)
            if self.validation_data is not None:
                self.validation_writer = tf.summary.FileWriter(os.path.join(self.p.main_path, "validation"), self.sess.graph)
            self.saver = tf.train.Saver(max_to_keep=100)

            # load Checkpoint
            if checkpoint_path:
                self.saver.restore(self.sess, checkpoint_path)

    def _encode(self, inputs):
        with tf.name_scope("encoder"):
            if self.p.periodicity < float("inf"):
                current_layer = tf.concat([tf.sin(inputs), tf.cos(inputs)], 1)
            else:
                current_layer = inputs

            assert len(self.p.n_neurons) == len(self.p.activation_functions) - 1, \
                "you need one activation function more then layers given in n_neurons"
            for i, (n_neurons, act_fun) in enumerate(zip(self.p.n_neurons, self.p.activation_functions[1:])):
                if act_fun:
                    act_fun = getattr(tf.nn, act_fun)
                else:
                    act_fun = None
                variable_summaries("activation{}".format(i), current_layer)
                dense = tf.layers.Dense(n_neurons, activation=act_fun,
                                        kernel_initializer=tf.variance_scaling_initializer,
                                        kernel_regularizer=self.regularizer,
                                        bias_initializer=tf.random_normal_initializer(0.1, 0.05))
                current_layer = dense(current_layer)
                add_layer_summaries(dense)
            return current_layer

    def _generate(self, inputs):
        with tf.name_scope("generator"):
            current_layer = inputs
            if self.p.periodicity < float("inf"):
                n_neurons_with_inputs = [self.train_data.shape[1] * 2] + self.p.n_neurons
            else:
                n_neurons_with_inputs = [self.train_data.shape[1]] + self.p.n_neurons
            for n_neurons, act_fun in zip(n_neurons_with_inputs[-2::-1], self.p.activation_functions[-2::-1]):
                if act_fun:
                    act_fun = getattr(tf.nn, act_fun)
                else:
                    act_fun = None
                current_layer = tf.layers.dense(current_layer, n_neurons,
                                                activation=act_fun,
                                                kernel_initializer=tf.variance_scaling_initializer,
                                                kernel_regularizer=self.regularizer,
                                                bias_initializer=tf.random_normal_initializer(0.1, 0.05))
            if self.p.periodicity < float("inf"):
                split = self.train_data.shape[1]
                current_layer = tf.atan2(current_layer[:, :split], current_layer[:, split:])
            return current_layer

    def _cost(self):
        self.auto_cost = tf.reduce_mean(
            tf.norm(periodic_distance(self.inputs, self.generated, self.p.periodicity), axis=1))
        self.distance_cost = distance_cost(self.inputs, self.latent, *self.p.dist_sig_parameters, self.p.periodicity)
        self.center_cost = tf.reduce_mean(tf.square(tf.norm(self.latent, axis=1)))
        # Todo: square of square root is inefficient
        self.reg_cost = tf.losses.get_regularization_loss()
        cost = self.p.distance_cost_scale * self.distance_cost \
               + self.p.auto_cost_scale * self.auto_cost \
               + self.p.center_cost_scale * self.center_cost \
               + self.reg_cost

        tf.summary.scalar("cost", cost)
        tf.summary.scalar("auto_cost", self.auto_cost)
        tf.summary.scalar("distance_cost", self.distance_cost)
        tf.summary.scalar("center_cost", self.center_cost)
        tf.summary.scalar("reg_cost", self.reg_cost)
        return cost

    def encode(self, inputs):
        """
        Projects high dimensional data to a low dimensional space using the encoder part of the autoencoder.

        :param inputs: 2d numpy array with the same number of columns as the used train_data
        :return: 2d numpy array with the point projected the the low dimensional space. The number of columns is equal
                 to the number of neurons in the bottleneck layer of the autoencoder.
        """
        latents = []
        batches = np.array_split(inputs, max(1, int(len(inputs) / 2048)))
        for batch in batches:
            latent = self.sess.run(self.latent, feed_dict={self.inputs: batch})
            latents.append(latent)
        latents = np.concatenate(latents, axis=0)
        return latents

    def generate(self, latent):
        """
        Generates new high-dimensional points based on given low-dimensional points using the decoder part of the
        autoencoder.

        :param latent: 2d numpy array containing points in the low-dimensional space. The number of columns must be
                       equal to the number of neurons in the bottleneck layer of the autoencoder.
        :return: 2d numpy array containing points in the high-dimensional space.
        """
        generateds = []
        batches = np.array_split(latent, max(1, int(len(latent) / 2048)))
        for batch in batches:
            generated = self.sess.run(self.generated, feed_dict={self.latent: batch})
            generateds.append(generated)
        generated = np.concatenate(generateds, axis=0)
        return generated

    def _random_batch(self, data, batch_size=None):
        batch_size = batch_size or self.p.batch_size
        batch = data[np.random.choice(len(data), batch_size, replace=False)]
        return batch

    def train(self):
        """
        Train the autoencoder as specified in the parameters object.
        """
        step = 0
        for step in tqdm(range(self.p.n_steps)):
            # feed_dict = {self.inputs: self.random_batch()}

            if step % self.p.summary_step == 0:
                _, summary_values = self.sess.run((self.optimize, self.merged_summaries))
                self.train_writer.add_summary(summary_values, step)
                if self.validation_data is not None:
                    summary_values = self.sess.run(self.merged_summaries,
                                                   feed_dict={self.inputs: self._random_batch(self.validation_data)})
                    self.validation_writer.add_summary(summary_values, step)
            else:
                self.sess.run(self.optimize)

            if step % self.p.checkpoint_step == 0 and step != 0:
                self.saver.save(self.sess, os.path.join(self.p.main_path, "checkpoints", "step{}.ckpt".format(step)))
        else:
            self.saver.save(self.sess, os.path.join(self.p.main_path, "checkpoints", "step{}.ckpt".format(step)))

    def close(self):
        """
        Close tensorflow session to free resources.
        :return:
        """
        self.sess.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from math import pi
import MDAnalysis as md
import numpy as np
from MDAnalysis.coordinates.memory import MemoryReader
from MDAnalysis.analysis.base import AnalysisFromFunction


def _expand_universe(universe, length):
    coordinates = AnalysisFromFunction(lambda ag: ag.positions.copy(),
                                       universe.atoms).run().results
    coordinates = np.tile(coordinates, (length, 1, 1))
    universe.load_new(coordinates, format=MemoryReader)


def _set_dihedral(dihedral, atoms, angle):
    current_angle = dihedral.dihedral.value()
    head = atoms[dihedral[2].id:]
    vec = dihedral[2].position - dihedral[1].position
    head.rotateby(angle-current_angle, vec, dihedral[2].position)


def dihedral_backmapping(pdb_path, dihedral_trajectory, rough_n_points=-1):
    """
    Takes a pdb file with a peptide and creates a trajectory based on the dihedral angles given.
    It simply rotates around the dihedral angle axis. In the result side-chains might overlap but the backbone should
    turn out quite well.

    :param pdb_path: (str)
    :param dihedral_trajectory:
        array-like of shape (traj_length, number_of_dihedrals)
    :param rough_n_points: (int) a step_size to select a subset of values from dihedral_trajectory is calculated by
        max(1, int(len(dihedral_trajectory) / rough_n_points)) with rough_n_points = -1 all values are used.
    :return: (MDAnalysis.Universe)
    """
    step_size = max(1, int(len(dihedral_trajectory) / rough_n_points))
    dihedral_trajectory = dihedral_trajectory[::step_size]

    uni = md.Universe(pdb_path)
    protein = uni.select_atoms("protein")

    dihedrals = []

    for residue in protein.residues:
        phi = residue.phi_selection()
        if phi:
            dihedrals.append(phi)

    for residue in protein.residues:
        psi = residue.psi_selection()
        if psi:
            dihedrals.append(psi)

    _expand_universe(uni, len(dihedral_trajectory))

    for dihedral_values, step in zip(dihedral_trajectory, uni.trajectory):
        for dihedral, value in zip(dihedrals, dihedral_values):
            _set_dihedral(dihedral, protein, value / (2 * pi) * 360)
    return uni



import os
from math import pi
import json
from .misc import search_and_replace, run_path


class ParametersFramework:
    def __init__(self):
        self.main_path = os.getcwd()

    def save(self, path=None):
        """
        Save parameters in json format

        :param path: Path where parameters should be saved. If no path is given main_path/parameters.json is used.
        :return: The path where the parameters were saved.
        """
        if not path:
            path = os.path.join(self.main_path, "parameters.json")
        with open(path, 'w') as file:
            json.dump(self.__dict__, file, indent=4, sort_keys=True)
        return path

    @classmethod
    def load(cls, path):
        """
        Loads the parameters saved in a json file into a new Parameter object.

        :param path: path of the json parameter file
        :return: a Parameter object
        """
        with open(path, 'r') as file:
            params = json.load(file)
        if params["main_path"] != os.path.dirname(path):
            print("seams like the directory was moved. Parameter file is updated ...")
            search_and_replace(path, params["main_path"], os.path.dirname(path))
            with open(path, 'r') as file:
                params = json.load(file)

        param = cls()
        param._setattrs(params)
        return param

    def _setattrs(self, dictionary):
        for key, value in dictionary.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise ValueError("The parameter " + key + " is not supported")

    def __getitem__(self, item):
        return getattr(self, item)


class Parameters(ParametersFramework):
    """
    :ivar main_path: Defines a main path where the parameters and other things might be stored.
    :ivar n_neurons: List containing number of neurons for each layer up to the bottleneck layer.
        For example [128, 128, 2] stands for an autoencoder with the following architecture
        {i, 128, 128, 2, 128, 128, i} where i is the number of dimensions of the input data.
    :ivar activation_functions: List of activation function names as implemented in TensorFlow.
        For example: "relu", "tanh", "sigmoid" or "" to use no activation function.
        The encoder part of the network takes the activation functions from the list starting with the second element.
        The decoder part of the network takes the activation functions in reversed order starting with
        the second element form the back. For example ["", "relu", "tanh", ""] would result in a autoencoder with
        {"relu", "tanh", "", "tanh", "relu", ""} as sequence of activation functions.
    :ivar periodicity: Defines the distance between periodic walls for the inputs.
        For example 2pi for angular values in radians.
        Set it to float("inf") for non-periodic inputs.

    :ivar learning_rate: Learning rate used by the optimizer.
    :ivar n_steps: Number of training steps.
    :ivar batch_size: Number of training points used in each training step
    :ivar summary_step: A summary for TensorBoard is writen every summary_step steps.
    :ivar checkpoint_step: A checkpoint is writen every checkpoint_step steps.

    :ivar dist_sig_parameters: Parameters for the sigmoid functions applied to the high- and low-dimensional distances
        in the following order (sig_h, a_h, b_h, sig_l, a_l, b_l)
    :ivar distance_cost_scale: Adjusts how much the distance based metric is weighted in the cost function.
    :ivar auto_cost_scale: Adjusts how much the autoencoding cost is weighted in the cost function.
    :ivar center_cost_scale: Adjusts how much the centering cost is weighted in the cost function.
    :ivar l2_reg_constant: Adjusts how much the l2 regularisation is weighted in the cost function.

    :ivar gpu_memory_fraction: Specifies the fraction of gpu memory blocked.
        This can be reduced if multiple runs should run on one gpu in parallel.
    :ivar analysis_path: A path that can be used to store analysis
    :ivar id: Can be any name for the run. Might be useful for example for specific analysis for different data sets.


    """
    def __init__(self):
        super(Parameters, self).__init__()
        self.n_neurons = [128, 128, 2]
        self.activation_functions = ["", "tanh", "tanh", ""]
        self.periodicity = 2 * pi

        self.learning_rate = 0.001
        self.n_steps = 100000
        self.batch_size = 256
        self.summary_step = 100
        self.checkpoint_step = 5000

        self.dist_sig_parameters = (4.5, 12, 6, 1, 2, 6)
        self.distance_cost_scale = 500
        self.auto_cost_scale = 1
        self.center_cost_scale = 0.0001
        self.l2_reg_constant = 0.001

        self.gpu_memory_fraction = 1
        self.analysis_path = ""
        self.id = ""



import numpy as np
from scipy.interpolate import interp1d
from matplotlib.lines import Line2D
from itertools import cycle
from matplotlib.widgets import Lasso
import os
from .misc import create_dir, periodic_distance_np, sigmoid
import MDAnalysis as md
import datetime
from .dihedral_backmapping import dihedral_backmapping
import matplotlib.pyplot as plt


class ManualPath(object):
    """
    ManualPath is a tool to manually select a path in a matplotlib graph.
    It supports two modes: "interpolated line", and "free draw".
    Press "m" to switch modes.

    In interpolated line mode klick in the graph to add an additional way point.
    Press "delete" to remove the last way point.
    Press "d" to remove all way points.
    Press "enter" once you have finished you path selection.

    In free draw mode press and hold the left mouse button while you draw a path.

    Once the path selection is completed, the use_points method is called with the points on the selected path.
    You can overwrite the use_points method to do what ever you want with the points on the path.
    """
    def __init__(self, axe, n_points=200):
        """

        :param axe: matplotlib axe object for example from: fig, axe = plt.subplots()
        :param n_points: Number of points distributed on the selected path.
        """
        self.axe = axe
        self.canvas = axe.figure.canvas
        self.fig = axe.figure
        self.lasso = None
        self.n_interpolation_points = n_points

        self.background = self.fig.canvas.copy_from_bbox(self.fig.bbox)

        control_line = Line2D([], [], ls='--', c='#666666',
                              marker='x', mew=2, mec='#204a87')
        self.control_line = self.axe.add_line(control_line)
        self.x_control = list(self.control_line.get_xdata())
        self.y_control = list(self.control_line.get_ydata())

        interpolated_line = Line2D([], [], c=self.control_line.get_markeredgecolor())
        self.interpolated_line = self.axe.add_line(interpolated_line)

        self.modes = cycle([self._add_point_interp, self._free_draw])
        self.mode = next(self.modes)
        self.click_cid = self.canvas.mpl_connect('key_press_event', self._on_key)
        self.click_cid = self.canvas.mpl_connect('button_press_event', self._on_click)
        self.draw_cid = self.canvas.mpl_connect('draw_event', self._grab_background)

    def _on_click(self, event):
        # Ignore clicks outside axes
        if event.inaxes != self.axe:
            return
        if self.canvas.widgetlock.locked():
            return
        if event.inaxes is None:
            return
        if event.button == 1:
            self.mode(event)

    def _on_key(self, event):
        if event.key == "m":
            self.mode = next(self.modes)
            return
        if event.key == "enter":
            points = np.array(self.interpolated_line.get_data()).T
            self._reset_lines()
            self.use_points(points)
            self._grab_background()

            return
        if event.key == "d":
            self._reset_lines()
            return
        if event.key == "delete":
            del self.x_control[-1]
            del self.y_control[-1]
            self._update_interp()

    def _free_draw(self, event):
        self.lasso = Lasso(event.inaxes,
                           (event.xdata, event.ydata),
                           self._free_draw_callback)
        # acquire a lock on the widget drawing
        self.canvas.widgetlock(self.lasso)

    def _free_draw_callback(self, verts):
        points = np.array(verts)
        self.use_points(points)
        self.canvas.draw_idle()
        self.canvas.widgetlock.release(self.lasso)
        del self.lasso

    def _add_point_interp(self, event):
        self.x_control.append(event.xdata)
        self.y_control.append(event.ydata)
        self._update_interp()

    def _update_interp(self):
        self.control_line.set_data(self.x_control, self.y_control)
        x_i, y_i = self._interpolate(self.x_control, self.y_control)
        x_i, y_i = self._interpolate(x_i, y_i)  # second iteration makes points more evenly spaced
        self.interpolated_line.set_data(x_i, y_i)
        self._update_lines()

    def _interpolate(self, x, y):
        cumulative_distances = [0]
        for i in range(1, len(x)):
            dist = ((x[i] - x[i - 1]) ** 2 + (y[i] - y[i - 1]) ** 2) ** 0.5
            cumulative_distances.append(cumulative_distances[-1] + dist)
        interp_i = np.linspace(0, max(cumulative_distances), self.n_interpolation_points)
        try:
            x_i = interp1d(cumulative_distances, x, kind='cubic')(interp_i)
            y_i = interp1d(cumulative_distances, y, kind='cubic')(interp_i)
        except ValueError:
            try:
                x_i = interp1d(cumulative_distances, x, kind='linear')(interp_i)
                y_i = interp1d(cumulative_distances, y, kind='linear')(interp_i)
            except ValueError:
                x_i = []
                y_i = []
        return x_i, y_i

    def use_points(self, points):
        """
        Overwrite this method to use the selected points in any way you like.

        For Example:

        >>> class MyManualPath(ManualPath):
        >>>     def use_points(self, points):
        >>>         print(points)

        :param points: numpy array with points from the manual path selection
        :return: None
        """
        self.axe.plot(points[:, 0], points[:, 1], linestyle="", marker=".")
        raise NotImplementedError("has to be implemented in subclass")

    def _grab_background(self, event=None):
        """
        When the figure is resized, hide the points, draw everything,
        and update the background.
        """
        # Thanks to: https://stackoverflow.com/questions/29277080/efficient-matplotlib-redrawing
        self.canvas.mpl_disconnect(self.draw_cid)
        self.interpolated_line.set_visible(False)
        self.control_line.set_visible(False)
        self.canvas.draw()

        # With most backends (e.g. TkAgg), we could grab (and refresh, in
        # self.blit) self.ax.bbox instead of self.fig.bbox, but Qt4Agg, and
        # some others, requires us to update the _full_ canvas, instead.
        self.background = self.fig.canvas.copy_from_bbox(self.fig.bbox)

        self.interpolated_line.set_visible(True)
        self.control_line.set_visible(True)
        self.canvas.draw()
        self.draw_cid = self.canvas.mpl_connect('draw_event', self._grab_background)

    def _update_lines(self):
        """
        Efficiently update the figure, without needing to redraw the
        "background" artists.
        """
        self.fig.canvas.restore_region(self.background)
        self.axe.draw_artist(self.interpolated_line)
        self.axe.draw_artist(self.control_line)
        self.canvas.update()
        self.canvas.flush_events()

    def _reset_lines(self):
        self.interpolated_line.set_data([[], []])
        self.control_line.set_data([[], []])
        self._update_lines()

        self.x_control = []
        self.y_control = []


class PathGenerateDihedrals(ManualPath):
    """
    This class inherits from :py:class:`encodermap.plot.ManualPath`.
    The points from a manually selected path are fed into the decoder part of a given autoencoder.
    The output of the autoencoder is used as phi psi dihedral angles to reconstruct protein conformations
    based on the protein structure given with pdb_path.
    Three output files are written for each selected path:
    points.npy, generated.npy and generated.pdb which contain:
    the points on the selected path, the generated output of
    the autoencoder, and the generated protein conformations respectively.
    Keep in mind that backbone dihedrals are not sufficient to describe a protein conformation completely.
    Usually the backbone is reconstructed well but all side chains are messed up.
    """

    def __init__(self, axe, autoencoder, pdb_path, save_path=None, n_points=200):
        """

        :param axe: matplotlib axe object for example from: fig, axe = plt.subplots()
        :param autoencoder: :py:class:`encodermap.autoencoder.Autoencoder` which was trained on protein dihedral
            angles. The dihedrals have to be order starting from the amino end.
            First all phi angles then all psi angles.
        :param pdb_path: Path to a protein data bank (pdb) file of the protein
        :param save_path: Path where outputs should be written
        :param n_points: Number of points distributed on the selected path.
        """
        super(PathGenerateDihedrals, self).__init__(axe, n_points=n_points)

        self.autoencoder = autoencoder
        self.pdb_path = pdb_path

        if save_path:
            self.save_path = save_path
        else:
            self.save_path = autoencoder.p.main_path

    def use_points(self, points):
        current_save_path = create_dir(os.path.join(self.save_path, "generated_paths",
                                                    datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")))
        self.axe.plot(points[:, 0], points[:, 1], linestyle="", marker=".")
        np.save(os.path.join(current_save_path, "points"), points)
        generated = self.autoencoder.generate(points)
        np.save(os.path.join(current_save_path, "generated.npy"), generated)

        universe = dihedral_backmapping(self.pdb_path, generated)
        output_pdb_path = os.path.join(current_save_path, "generated.pdb")
        with md.Writer(output_pdb_path) as w:
            for step in universe.trajectory:
                w.write(universe.atoms)


def distance_histogram(data, periodicity, sigmoid_parameters, axe=None):
    """
    Plots the histogram of all pairwise distances in the data.
    If sigmoid parameters are given it also shows the sigmoid function and its normalized derivative.

    :param data: each row should contain a point in a number_of _columns dimensional space.
    :param periodicity: Periodicity of the data. use float("inf") for non periodic data
    :param sigmoid_parameters: tuple (sigma, a, b)
    :param axe: matplotlib axe object ore None. If None a new figure is generated.
    :return:
    """
    vecs = periodic_distance_np(np.expand_dims(data, axis=1), np.expand_dims(data, axis=0), periodicity)
    dists = np.linalg.norm(vecs, axis=2)
    dists = dists.reshape(-1)

    if axe is None:
        fig, axe = plt.subplots()
    axe2 = axe.twinx()
    axe2.hist(dists, bins="auto", density=True)
    x = np.linspace(0, max(dists), 1000)

    y = sigmoid(x, *sigmoid_parameters)
    dy = np.diff(y)
    dy_norm = dy / max(dy)
    axe.plot(x, y, color="C1", label="sigmoid")
    axe.plot(x[:-1], dy_norm, color="C2", label="diff sigmoid")

    axe.legend()
    axe.set_xlabel("distance")
    axe.set_ylim((0, 1))
    axe.set_zorder(axe2.get_zorder() + 1)
    axe.patch.set_visible(False)
    return axe



from setuptools import setup

with open("description.md", "r") as fh:
    long_description = fh.read()

setup(name='encodermap',
      version='0.0.5',
      description='python library for dimensionality reduction',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Tobias Lemke',
      url="https://github.com/AG-Peter/encodermap",
      packages=['encodermap'],
      install_requires=[
          'numpy',
          'matplotlib',
          'scipy',
          'MDAnalysis',
          'tqdm'
      ],
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Operating System :: OS Independent",
      ])



import tensorflow as tf
from math import pi
import os
import numpy as np


def add_layer_summaries(layer):
    """

    :param layer:
    :return:
    """
    weights = layer.variables[0]
    biases = layer.variables[1]
    variable_summaries(layer.name + "_weights", weights)
    variable_summaries(layer.name + "_biases", biases)


def periodic_distance(a, b, periodicity=2*pi):
    """

    :param a:
    :param b:
    :param periodicity:
    :return:
    """
    d = tf.abs(b-a)
    return tf.minimum(d, periodicity-d)


def periodic_distance_np(a, b, periodicity=2*pi):
    """

    :param a:
    :param b:
    :param periodicity:
    :return:
    """
    d = np.abs(b-a)
    return np.minimum(d, periodicity-d)


def variable_summaries(name, variables):
    """
    Attach several summaries to a Tensor for TensorBoard visualization.

    :param name:
    :param variables:
    :return:
    """
    if not isinstance(variables, list):
        variables = [variables]

    for i, var in enumerate(variables):
        try:
            add_index = len(variables) > 1
        except TypeError:
            add_index = True
        if add_index:
            name = name + str(i)
        with tf.name_scope(name):
            mean = tf.reduce_mean(var)
            tf.summary.scalar('mean', mean)
            with tf.name_scope('stddev'):
                stddev = tf.sqrt(tf.reduce_mean(tf.square(var - mean)))
            tf.summary.scalar('stddev', stddev)
            tf.summary.scalar('max', tf.reduce_max(var))
            tf.summary.scalar('min', tf.reduce_min(var))
            tf.summary.histogram('histogram', var)


def create_dir(path):
    """

    :param path:
    :return:
    """
    if not os.path.isdir(path):
        os.makedirs(path)
    return path


def distance_cost(r_h, r_l, sig_h, a_h, b_h, sig_l, a_l, b_l, periodicity):
    """

    :param r_h:
    :param r_l:
    :param sig_h:
    :param a_h:
    :param b_h:
    :param sig_l:
    :param a_l:
    :param b_l:
    :param periodicity:
    :return:
    """

    dist_h = pairwise_dist_periodic(r_h, periodicity)
    dist_l = pairwise_dist(r_l)

    sig_h = sigmoid(dist_h, sig_h, a_h, b_h)
    sig_l = sigmoid(dist_l, sig_l, a_l, b_l)

    cost = tf.reduce_mean(tf.square(sig_h - sig_l))
    return cost


def sigmoid(r, sig, a, b):
    """

    :param r:
    :param sig:
    :param a:
    :param b:
    :return:
    """
    return 1 - (1 + (2**(a/b) - 1) * (r/sig)**a)**(-b/a)


def pairwise_dist_periodic(positions, periodicity):
    vecs = periodic_distance(tf.expand_dims(positions, axis=1), tf.expand_dims(positions, axis=0), periodicity)
    mask = tf.to_float(tf.equal(vecs, 0.0))
    vecs = vecs + mask * 1e-16  # gradient infinite for 0
    dists = tf.norm(vecs, axis=2)
    return dists


def pairwise_dist(positions, squared=False):
    """
    Compute the 2D matrix of distances between all the embeddings.

    :param positions: tensor of shape (batch_size, position_dim)
    :param squared: Boolean. If true, output is the pairwise squared euclidean distance matrix.
        If false, output is the pairwise euclidean distance matrix.

    :return: the pairwise_distances as a tensor of shape (batch_size, batch_size)
    """
    # thanks to https://omoindrot.github.io/triplet-loss

    # Get the dot product between all embeddings
    # shape (batch_size, batch_size)
    dot_product = tf.matmul(positions, tf.transpose(positions))

    # Get squared L2 norm for each embedding. We can just take the diagonal of `dot_product`.
    # This also provides more numerical stability (the diagonal of the result will be exactly 0).
    # shape (batch_size,)
    square_norm = tf.diag_part(dot_product)

    # Compute the pairwise distance matrix as we have:
    # ||a - b||^2 = ||a||^2  - 2 <a, b> + ||b||^2
    # shape (batch_size, batch_size)
    distances = tf.expand_dims(square_norm, 0) - 2.0 * dot_product + tf.expand_dims(square_norm, 1)

    # Because of computation errors, some distances might be negative so we put everything >= 0.0
    distances = tf.maximum(distances, 0.0)

    if not squared:
        # Because the gradient of sqrt is infinite when distances == 0.0 (ex: on the diagonal)
        # we need to add a small epsilon where distances == 0.0
        mask = tf.to_float(tf.equal(distances, 0.0))
        distances = distances + mask * 1e-16

        distances = tf.sqrt(distances)

        # Correct the epsilon added: set the distances on the mask to be exactly 0.0
        distances = distances * (1.0 - mask)

    return distances


def search_and_replace(file_path, search_pattern, replacement, out_path=None, backup=True):
    """
    Searches for a pattern in a text file and replaces it with the replacement

    :param file_path: (str)
            path to the file to search
    :param search_pattern: (str)
            pattern to search for
    :param replacement: (str)
            string that replaces the search_pattern in the output file
    :param out_path: (str)
            path where to write the output file. If no path is given the original file will be replaced.
    :param backup: (bool)
            if backup is true the original file is renamed to filename.bak before it is overwritten

    :return:

    """
    with open(file_path, "r") as f:
        file_data = f.read()

    file_data = file_data.replace(search_pattern, replacement)

    if not out_path:
        out_path = file_path
        if backup:
            os.rename(file_path, file_path+".bak")

    with open(out_path, "w") as file:
        file.write(file_data)


def run_path(path):
    """
    Creates a directory at "path/run{i}" where i the i corresponding to the smallest not yet existing path

    :param path: (str)
    :return: (str)
        path of the created folder

    """
    i = 0
    while True:
        current_path = os.path.join(path, "run{}".format(i))
        if not os.path.exists(current_path):
            os.makedirs(current_path)
            output_path = current_path
            break
        else:
            i += 1
    return output_path


def random_on_cube_edges(n_points, sigma=0):
    x = y = z = 1
    r = np.random.uniform(size=n_points)

    coordinates = np.zeros((n_points, 3))

    ids = np.zeros(n_points)

    a = np.array([[0, 0, 0]]*3 +
                 [[x, y, 0]]*3 +
                 [[0, y, z]]*3 +
                 [[x, 0, z]]*3)

    b = np.array([[x, 0, 0],
                  [0, y, 0],
                  [0, 0, z],
                  [-x, 0, 0],
                  [0, -y, 0],
                  [0, 0, z],
                  [x, 0, 0],
                  [0, -y, 0],
                  [0, 0, -z],
                  [-x, 0, 0],
                  [0, y, 0],
                  [0, 0, -z],
                  ])

    for i in range(12):
        mask = (i/12 < r) & (r < (i+1)/12)
        coordinates[mask] += a[i] + (np.expand_dims(r[mask], axis=1) - i/12) * 12 * b[i]
        ids[mask] = i

    if sigma:
        coordinates += np.random.normal(scale=sigma, size=(n_points, 3))

    return coordinates, ids
